unit ExProgressbar;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Graphics, Types, Controls, ExtCtrls, ComCtrls;

type
  TProgressbarEx = class(TGraphicControl)
  private
    const
      DEFAULT_BACKCOLOR = $E6E6E6;
      DEFAULT_BARCOLOR1 = $06B025;
      DEFAULT_BARCOLOR2 = $43BB28;
      DEFAULT_BORDERCOLOR = clSilver;
      DEFAULT_MARQUEELENGTH = 120;
  private
    FCaption: String;
    FColors: array[0..3] of TColor;
    FMarqueePosition: Integer;
    FMarqueeLength: Integer;
    FMarqueeSpeed: Integer;
    FMax: Integer;
    FMin: Integer;
    FOrientation: TProgressBarOrientation; // = (pbHorizontal, pbVertical, pbRightToLeft, pbTopDown);
    FPosition: Integer;
    FStyle: TProgressBarStyle; // = (pbstNormal, pbstMarquee);
    FTimer: TTimer;
    function GetColors(AIndex: Integer): TColor;
    procedure SetCaption(AValue: String);
    procedure SetColors(AIndex: Integer; AValue: TColor);
    procedure SetMax(AValue: Integer);
    procedure SetMin(AValue: Integer);
    procedure SetOrientation(AValue: TProgressbarOrientation);
    procedure SetPosition(AValue: Integer);
    procedure SetStyle(AValue: TProgressbarStyle);
    procedure TimerHandler(Sender: TObject);
  protected
    class function GetControlClassDefaultSize: TSize; override;
    function IsHorizontal(AOrientation: TProgressbarOrientation): Boolean;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property BackColor: TColor index 0 read GetColors write SetColors default DEFAULT_BACKCOLOR;
    property BarColor1: TColor index 1 read GetColors write SetColors default DEFAULT_BARCOLOR1;
    property BarColor2: TColor index 2 read GetColors write SetColors default DEFAULT_BARCOLOR2;
    property BorderColor: TColor index 3 read GetColors write SetColors default DEFAULT_BORDERCOLOR;
    property Caption: String read FCaption write SetCaption;
    property MarqueeLength: Integer read FMarqueeLength write FMarqueeLength default DEFAULT_MARQUEELENGTH;
    property MarqueeSpeed: Integer read FMarqueeSpeed write FMarqueeSpeed default 8;
    property Max: Integer read FMax write SetMax default 100;
    property Min: Integer read FMin write SetMin default 0;
    property Orientation: TProgressbarOrientation read FOrientation write SetOrientation default pbHorizontal;
    property Position: Integer read FPosition write SetPosition default 0;
    property Style: TProgressbarStyle read FStyle write SetStyle default pbstNormal;

    property Align;
    property Anchors;
    property Font;
  end;

implementation

constructor TProgressbarEx.Create(AOwner: TComponent);
begin
  inherited;
  FCaption := '';
  FColors[0] := DEFAULT_BACKCOLOR;
  FColors[1] := DEFAULT_BARCOLOR1;
  FColors[2] := DEFAULT_BARCOLOR2;
  FColors[3] := DEFAULT_BORDERCOLOR;
  FMarqueeLength := DEFAULT_MARQUEELENGTH;
  FMarqueeSpeed := 8;
  FMax := 100;
  FStyle := pbstNormal;
  FTimer := TTimer.Create(self);
  FTimer.Interval := 25;
  FTimer.OnTimer := @TimerHandler;
  with GetControlClassDefaultSize do
    SetInitialBounds(0, 0, CX, CY);
end;

function TProgressbarEx.GetColors(AIndex: Integer): TColor;
begin
  Result := FColors[AIndex];
end;

class function TProgressbarEx.GetControlClassDefaultSize: TSize;
begin
  Result.CX := 200;
  Result.CY := 20;
end;

function TProgressbarEx.IsHorizontal(AOrientation: TProgressbarOrientation): Boolean;
begin
  Result := (AOrientation in [pbHorizontal, pbRightToLeft]);
end;

procedure TProgressbarEx.Paint;
var
  R: TRect;
  fraction: Double;
  FontAngle: Integer;
  txtSize: TSize;
  isHor: Boolean;
begin
  isHor := IsHorizontal(FOrientation);

  // Draw background with border
  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := FColors[0];
  Canvas.Pen.Color := FColors[3];
  Canvas.Rectangle(0, 0, Width, Height);

  // Draw bar
  fraction := (FPosition - FMin) / (FMax - FMin);
  R := Rect(1, 1, Width-1, Height-1);
  case FStyle of
    pbstNormal:
      begin
        if FMax = FMin then
          exit;
        case FOrientation of
          pbHorizontal: R.Right := round(fraction * (Width - 2));
          pbRightToLeft: R.Left := R.Right - round(fraction * (Width - 2));
          pbVertical: R.Top := R.Bottom - round(fraction * (Height - 2));
          pbTopDown: R.Bottom := round(fraction * (Height - 2));
        end;
      end;
    pbstMarquee:
      case FOrientation of
        pbHorizontal:
          begin
            R.Right := FMarqueePosition;
            R.Left := R.Right - FMarqueeLength;
          end;
        pbRightToLeft:
          begin
            R.Left := R.Right - FMarqueePosition;
            R.Right := R.Left + FMarqueelength;
          end;
        pbVertical:
          begin
            R.Top := R.Bottom - FMarqueePosition;
            R.Bottom := R.Top + FMarqueeLength;
          end;
        pbTopDown:
          begin
            R.Bottom := FMarqueePosition;
            R.Top := R.Bottom - FMarqueeLength;
          end;
      end;
  end;
  if isHor then
  begin
    if R.Right >= Width - 1 then R.Right := Width - 1;
    if R.Left < 1 then R.Left := 1;
    FontAngle := 0;
  end else
  begin
    if R.Bottom >= Height - 1 then R.Bottom := Height - 1;
    if R.Top < 1 then R.Top := 1;
    FontAngle := 900;
  end;
  Canvas.GradientFill(R, FColors[1], FColors[2], gdVertical);

  // Draw text
  if FCaption <> '' then
  begin
    txtSize := Canvas.TextExtent(FCaption);
    Canvas.Font.Assign(Font);
    Canvas.Font.Orientation := fontAngle;
    Canvas.Brush.Style := bsClear;
    R := Rect(0, 0, Width, Height);
    if isHor then
      Canvas.TextOut((Width - txtSize.CX) div 2, (Height - txtSize.CY) div 2, FCaption)
    else
      Canvas.TextOut((Width - txtSize.CY) div 2, R.Bottom - (Height - txtSize.CX) div 2, FCaption);
  end;
end;

procedure TProgressbarEx.SetCaption(AValue: String);
begin
  if FCaption = AValue then
    exit;
  FCaption := AValue;
  Invalidate;
end;

procedure TProgressbarEx.SetColors(AIndex: Integer; AValue: TColor);
begin
  if FColors[AIndex] = AValue then
    exit;
  FColors[AIndex] := AValue;
  Invalidate;
end;

procedure TProgressbarEx.SetMax(AValue: Integer);
begin
  if FMax = AValue then
    exit;
  FMax := AValue;
  Invalidate;
end;

procedure TProgressbarEx.SetMin(AValue: Integer);
begin
  if FMin = AValue then
    exit;
  FMin := AValue;
  Invalidate;
end;

procedure TProgressbarEx.SetOrientation(AValue: TProgressbarOrientation);
var
  rotate: Boolean;
begin
  if FOrientation = AValue then
    exit;
  rotate := IsHorizontal(AValue) xor IsHorizontal(FOrientation);
  FOrientation := AValue;
  if rotate then
    SetBounds(Left, Top, Height, Width)
  else
    Invalidate;
end;

procedure TProgressbarEx.SetPosition(AValue: Integer);
begin
  if FPosition = AValue then
    exit;
  if AValue > FMax then
    FPosition := FMax
  else if AValue < FMin then
    FPosition := FMin
  else
    FPosition := AValue;
  Invalidate;
end;

procedure TProgressbarEx.SetStyle(AValue: TProgressbarStyle);
begin
  if FStyle = AValue then
    exit;
  FStyle := AValue;
  case FStyle of
    pbstNormal:
      begin
        FPosition := FMin;
        FTimer.Enabled := false;
      end;
    pbstMarquee:
      begin
        FMarqueePosition := 0;
        FTimer.Enabled := true;
      end;
  end;
end;

procedure TProgressbarEx.TimerHandler(Sender: TObject);
begin
  inc(FMarqueePosition, FMarqueeSpeed);
  if IsHorizontal(FOrientation) then
  begin
    if (FMarqueePosition - FMarqueeLength > Width) and (FMarqueeSpeed > 0) then
      FMarqueePosition := 0
    else
    if (FMarqueePosition < 0) and (FMarqueeSpeed < 0) then
      FMarqueePosition := Width + FMarqueeLength;
  end else
  begin
    if (FMarqueePosition - FMarqueeLength > Height) and (FMarqueeSpeed > 0) then
      FMarqueePosition := 0
    else
    if (FMarqueePosition < 0) and (FMarqueeSpeed < 0) then
      FMarqueePosition := Height + FMarqueeLength;
  end;
  Invalidate;
end;

end.

